.drivecode_include
!src "config.inc"
!convtab pet
!cpu 6510

;XXX TODO have sector loading in ZP instead of all this?

;--------------- ZP usage ----------------
;.. = unused
;bl = blocks on list
;ms = max sectors on current track
;fs = filesize in blocks (256 byte lock)
;tr = current track
;in = number of currently loaded block
;pl = preamble bytes lo nibbles
;ph = preamble bytes hi nibbles
;sh = sector header data (gcr encoded -> 5 lo- and hinibbles)
;ds = dirsect
;wi = windex, index to be used for in order loading
;bs = blocksize, size of transferred block

;00: .. .. .. .. .. .. .. ..
;08: bl 0e 0f 07 tr 0a 0b 03 <-+
;10: ms ds 0d 05 in 00 09 01 <-| gcr2ser table with gaps being used by other variables
;18: fs 06 0c 04 bs 02 08 wi <-+
;20: 0f 07 0d 05 0b 03 09 01 <-+ bin2ser lookup table
;28: 0e 06 0c 04 0a 02 08 00 <-+
;30: halftrack code ........
;38: is ....................
;40: here ...... .. .. .. ..
;48: 0e 08 80 sh sh sh sh sh
;50: 0f 00 00 sh sh sh sh sh
;58: 07 01 10 pl pl pl pl ..
;60: fi fi fi fi fi fi .. ..		;filestruct
;68: 0a 0c c0 ph ph ph ph ..
;70: 0b 04 40 .. .. .. .. ..
;78: 03 05 50 ff ff ff ff ff <-+
;80: ff ff ff ff ff ff ff ff <-| list of wanted sectors
;88: ff ff ff ff ff ff ff ff <-+
;90: 0d 02 20 .. .. .. .. ..
;98: 05 03 30 motor + led ..
;a0: code here ....... .. ..
;a8: 00 0f f0 .. .. .. .. ..
;b0: 09 06 60 .. .. .. .. ..
;b8: 01 07 70 turn disc ....
;c0: entry code here .......
;c8: 06 09 90 .. .. .. .. ..
;d0: 0c 0a a0 .. .. .. .. ..
;d8: 04 0b b0 .. .. .. .. ..
;e0: .. .. .. .. .. .. .. ..
;e8: 02 0d d0 .. .. .. .. ..
;f0: 08 0e e0 .. .. .. .. ..
;f8: .. .. .. .. .. .. .. ..
;    ^  ^  ^
;    |  |  |
;    |  |  gcr dec_hi with index << 3
;    |  gcr_dec_lo with index << 3
;    gcr2ser with index << 3
;
;thus it is sufficient to inflate the gcr values in that way that they are either right or left aligned
;-----------------------------------------

.STEPPING_SPEED	= $93

.drivecode	= $0108
.bootstrap	= $0700

;VIA1		= $1800
;VIA2		= $1c00

.VIA2_LED_OFF	= $f7
.VIA2_LED_ON	= $08

.VIA2_MOTOR_OFF	= $fb
.VIA2_MOTOR_ON	= $04

.blocks_on_list	= $08		;blocks tagged on wanted list
.track		= $0c		;current track
.max_sectors	= $10		;maximum sectors on current track
.dirsect	= $11
.index		= $14		;current blockindex
.blocks		= $18		;number of blocks the file occupies
.blocksize	= $1c
!if BITFIRE_CONFIG_IN_ORDER = 1 {
.windex		= $1f
}

.file_descriptor = $60
.to_track	= .file_descriptor + 0
.sector		= .file_descriptor + 1
.ld_addr	= .file_descriptor + 2
.file_size	= .file_descriptor + 4

.temp		= .sector
.cmp1		= .sector
.cmp2		= .index
.dest		= $52		;-> 52 is always zero for free $53 will be set, sectorheader is unused during upload

.halftrack_code	= $30
.motor_code	= $9b
.turn_code	= $bb

.preamble_lo	= $5b		;preamble data gcr coded lonibbles
.preamble_hi	= $6b		;preamble data gcr coded hinibbles

.wanted		= $7b		;list of wanted sectors/blockindices -> $15 bytes space

.head_lo	= $4b		;sector header gcr data lonibbles
.head_hi	= $53		;sector header gcr data hinibbles

.gcr2ser	= $00		;gcr to serial-port lookuptablea -> from $08-$1f
.bin2ser	= $20		;binary to serial data

.directory	= $0500		;directory
.lonibbles	= $0600
.hinibbles	= $0700

.gcr_dec_lo	= $f8c0
.gcr_dec_hi	= $f8a0
.gcr_dec_lo_s	= $01
.gcr_dec_hi_s	= $02

.DIR_SECT	= 18

BUSY		= $02
BLOCK_READY	= $08
IDLE		= $00

.bootstrap_start
!pseudopc .bootstrap {
.bootstrap_run
		lda #$12
		sta $0a
		lda #.DIR_SECT
		sta $0b

		;lda #$c0
		;sta $02
		;lda $02
		;bmi *-2

		lda #$80
		sta $02
		lda $02
		bmi *-2

		lda $1c00
		and #(.VIA2_LED_OFF & .VIA2_MOTOR_OFF)
		sta $1c00

		;XXX TODO currtrack is in $18?
                sei

		ldx #<.drivecode-1
		txs
		lda #$00
-
		sta $0100,x
		dex
		bpl -

                lda #%01111010			;DDR set bits for drivenumber to 0, ATN out, CLK out and DATA out are outputs
                sta $1802
		lda #%11101110			;CA2 manual output high, CA1 low, CB2 manual output high, CB1 low (NC)
		sta $1c0c
		lda #%01000001
		sta $1c0b

		lda #$7f			;disable all interrupts
		sta $180e
		sta $1c0e

		lda #(>.drivecode_size)+1
		sta .temp

		ldy #$00
.get_block
		;signal that no block is ready yet, asap
		ldx #BLOCK_READY
		stx $1800

		;wait for atn
		ldx $1800
		bpl *-3

		;clear all lines and set bit 7 as bit counter, to allow data in and clk in to be set/cleared by host
		lda #$80
		sta $1800

		ldx $1800
-
		;did a bit arrive? (bit flip in data in, atn is dropped in same go in first bit)
		cpx $1800
		beq *-3

		;load register
		ldx $1800

		;push bit into carry
		cpx #$04

		;shift in
		ror

		;do until our counter bit ends up in carry
		bcc -

.block = * + 1
                sta .drivecode,y

                iny
                bne .get_block

                inc .block+1
                dec .temp
                bne .get_block

		lda #BUSY
		sta $1800
		jmp .drivecode_setup
}

.bootstrap_end
.bootstrap_size = .bootstrap_end - .bootstrap_start

.drivecode_start
!pseudopc .drivecode {
.drivecode_run
		ldx #.DIR_SECT
		stx .dirsect
		jsr .read_dir
		jmp .motor_off		;after that jmp idle
.idle
		;receive a byte
		jsr .get_byte
		;load file, file number is in A

.load_file
		cmp #$f0
		bcc *+5
		jmp .turn_disc
		tay
		bpl *+5
		jmp .upload

		ldx #.DIR_SECT+1
-
		dex
		sec
		sbc #42
		bcs -
		;underrun, add 42 again
		adc #42
		;store file number
		sta .temp
		;file is still in currently held dirsect?
		cpx .dirsect
		beq +
		;nope, advance
		stx .dirsect
		jsr .read_dir
+
		;file number * 6 to get proper index
		asl		;*2
		;clc		;can be omitted, not more than 42 files are allowed per side, so carry is always cleared after asl
		adc .temp	;+1 -> * 3
		asl		;*2 -> * 6
		tay

		;turn on led and motor, if not on already
		jsr .motor_on

		;reset block index
		ldx #$00
		stx .index
!if BITFIRE_CONFIG_IN_ORDER = 1 {
		stx .windex
}
-
		lda .directory,y
		sta .file_descriptor,x
		iny
		inx
		cpx #$06
		bne -

		;a = file_size highbyte, carry set
		adc #$00			;need to add one, as also the last partial block counts as a full block. File size is one too less, so exceptions like $xx00 in length are no problem
		sta .blocks

		;prepare load-address for first transferred block
		lda .ld_addr+1
		jsr .add_preamble
		lda .ld_addr
		jsr .add_preamble

		;to be entered with desired start sector (most likely 0 on a new track and start sector on the first track)
.load_track
		;go to track
		jsr .seek			;sets max_sectors and bitrate depending on track

		;now generate list of wanted blocks
		;clear block list counter and revolution counter Y
		lda #$00
		sta .blocks_on_list
		tay
-
		;desired sector?
		cmp .sector
		;yep, found it
		beq .hit
		;add interleave
		clc				;XXX always clear? omit?
		adc #BITFIRE_CONFIG_INTERLEAVE
		;take care about wrap around
		cmp .max_sectors
		bcc -

		iny				;no hit in this revolution, next one
		tya				;start next round with Y as offset
		bne -				;BRA
.hit
		;A is desired sector again, move over to X
--
		tax
-
		;get index
		lda .index
		;advance index
		inc .index
		;write index into wantedlist
		sta .wanted,x
		;count number of blocks in list
		inc .blocks_on_list
		;all blocks loaded?
		dec .blocks
		;yep, end
		beq .load_wanted_blocks
		;go for next block
		txa
		sbx #-BITFIRE_CONFIG_INTERLEAVE
		;wrap around?
		cpx .max_sectors
		bcc -
		;increase offset
		iny
		;start new round with new offset
		tya
		;offset > interleave?
		cpy #BITFIRE_CONFIG_INTERLEAVE
		bne --
		;yep, finish

.load_wanted_blocks
		;read and transfer all blocks on wishlist
		jsr .read_sector
		jsr .preamble_size

		;is it the last block of a track being transferred? then step a halftrack up first and then send data which is also our waiting loop to let the stepper settle
		dec .blocks_on_list
		bne +
		;but only do so if it is not the last block of our file!
		lda .blocks
		beq +
		;set timer
		lda #.STEPPING_SPEED
		sta $1c05

		;do halftrack up immediatedly
		lda #$00
		jsr .halftrack__
+
		jsr .send_block

		;last block on wishlist?
		;XXX this is tested twice, could be avoided?
		lda .blocks_on_list
		bne .load_wanted_blocks
.track_finished
		;set start pos for next track
		;lda #$00
		sta .sector

		;check if all blocks of file are loaded, load X to preserve A = 0
		ldx .blocks
		;done?
		beq .done

		;do another halftrack, but respect minimum wait timer
		jsr .halftrack_
		;full track stepped
		inc .track

		;now adjust to_track and take care of skipping track 18
		ldx .track
		;skip dirtrack however
		cpx #18
		bne +
		inx
+
		stx .to_track
		jmp .load_track

.turn_loop
		;fetch first dir sector
		ldx #.DIR_SECT
		stx .dirsect
		jsr .read_dir_

		;compare side info
		lda .temp
		cmp .directory + $ff
		;nope, not the requested side
		bne .turn_loop

		;disc turn okay
		lda #BLOCK_READY | BUSY
		sta $1800

		;wait for response of host
		bit $1800
		bpl *-3

		;idle
.done
		lda #IDLE
		sta $1800
.motor_off
		lda $1c00
!if BITFIRE_CONFIG_MOTOR_ALWAYS_ON = 0 {
		and #(.VIA2_LED_OFF & .VIA2_MOTOR_OFF)
} else {
		and #.VIA2_LED_OFF
}
		sta $1c00
		jmp .idle

.seek
		;how many tracks to go?
		lda .track
		sec
		sbc .to_track
		beq .set_bitrate

		;up or downwards?
		ldy #$01
		bcs .seek_down
.seek_up
		;up we go, so invert num tracks as it is negative
		eor #$ff
		adc #$01
		dey
.seek_down
		;counter is twice the number of tracks (halftracks)
		asl
		tax
		;skip waitloop for the first halftrack, as it is not needed, thus on single track changes we save half the waiting \o/
		;bne skip_wait			;BRA
		;XXX have to disable that optimization when doing halftracks around sending of last block of track, else when skipping track 18 we change tracks too fast and might miss tracks
.step
		lda #.STEPPING_SPEED
		sta $1c05
.skip_wait
		;do halftrack up/down and respect minimum wait timer
		jsr .halftrack
		dex
		bne .step
.set_bitrate
		ldy .to_track
		sty .track
		;sets bitrate for current track and also sets value for max. sectors on track
		ldx #$11
		lda $1c00
		and #$9f
		cpy #31
		bcs +
		inx
		adc #$20
		cpy #25
		bcs +
		inx
		adc #$20
		cpy #18
		bcs +
		inx
		inx
		adc #$20
+
		stx .max_sectors
		sta $1c00
		asr #$60
		lsr
		lsr
		lsr
		lsr
		sta .smc1+1
		sta .smc2+1
		;sta .smc3+1
		rts

.read_dir
		jsr .motor_on
.read_dir_
		inc .wanted,x
		lda #$12
		sta .to_track
		sta .file_size + 1	;write any number > 0 to file_size + 1 to trigger a sector_length of 256 bytes
		jsr .seek

		jsr .read_sector

		;XXX TODO this is very ugly!!! :-(
		iny
-
		ldx .lonibbles,y
		cpx #$20
		bcs +
		lda .gcr_dec_lo,x
		!byte $2c
+
		lda+1 .gcr_dec_lo_s,x

		ldx .hinibbles,y
		cpx #$20
		bcs +
		ora .gcr_dec_hi,x
		!byte $2c
+
		ora+1 .gcr_dec_hi_s,x

		sta $0500,y
		dey
		bne -
		rts

.read_sector
		jsr .wait_sync_mark
		;no need to hurry when loading the header, so we can waste cycles with a few more bvc *
					;11111 222	header type
		cpx #$52
		bne .read_sector
		clv
		bvc *

		lda $1c01		;22333334	type  / checksum
		asr #%00111111
		sta .head_hi+1		;store @ pos 1						;7

.chksum
		clv
		bvc *

		lax $1c01		;44445555	checksum / sector
		ror
		lsr
		lsr
		lsr
		sta .head_lo+1
		clv
					;15 cycles
		bvc *

		txa
		ldx $1c01		;56666677	sector / track
		cpx #$80
		rol
		and #%00011111
		sta .head_hi+0		;store @ pos 0 instead of pos 1 to reuse value later on
		clv
					;15 cycles
		bvc *

		txa
		asr #%01111111		;bit 4 of 7 in carry
		sta .head_lo+0		;1 bit of 7 left
		lax $1c01		;77788888	track
		ror			;-> 77778888
		lsr .head_lo+0		;fetch msb of 7

		;clv
		;bvc *

		ror			;-> 77777888
		lsr
		lsr
		lsr
		sta .head_hi+2		;-> ---77777
		lda #%00011111
		sax .head_lo+2		;-> ---88888
					;read after 40 cycles
		ldx #$07
		lda $1c01		;11111222	id2
		clv

		bvc *

		sax .head_lo+3
		lsr
		lsr
		lsr
		sta .head_hi+3		;---11111

		lax $1c01		;22333334	id2/id1
		asl
		rol .head_lo+3
		asl
		rol .head_lo+3
		txa
		asr #%00111111
					;read on cycle 40
		sta .head_hi+4
		lda $1c01		;44445555	id2
!if >* != >.chksum { !error "header-checksum not in one page: ", .chksum, " - ", * }
		ror
		lsr
		lsr
		lsr
		sta .head_lo+4

		ldx #$04
		;decode sector header
-
		ldy .head_hi,x
		lda .gcr_dec_hi,y
		ldy .head_lo,x
		ora .gcr_dec_lo,y
		sta .head_lo,x
		dex
		bpl -
		tax			;sector number

		;header checksum A = .head_lo + 0
		eor .head_lo + 1
		eor .head_lo + 2
		eor .head_lo + 3
		eor .head_lo + 4
		;urgh, bad checksum, find another sector
		bne .back

		;check track, as we could easily read from wrong track during settling time of stepper (happend so in the wild, what a bother)
		lda .head_lo + 2
		cmp .track
		bne .back

		;sector on list?
		ldy .wanted,x
!if BITFIRE_CONFIG_IN_ORDER = 1 {
		;next sector? nope?
		cpy .windex
		bne .back
}
		iny			;$ff + 1 = 0?
		;yes, read, we should still have ~20 cycles overhead here until next sync arrives
		bne .wait_sector_data
.back
		jmp .read_sector

.wait_sync_mark
		lda $1c00		;wait for end of sync
		bmi *-3
		;and #.VIA2_LED_OFF
		;sta $1c00
		lda $1c01		;sync mark
		clv
		bvc *
		lax $1c01		;11111222	must be $55, no need to check
		rts
.wait_sector_data
		jsr .wait_sync_mark
		cpx #$55
		bne .back
		clv
		bvc *
		lda $1c01		;22333334
		asr #%00111111
		clv
		bvc *
		ldy #$00		;waste 6 cycles to have more or less same entry point into loop
		nop
		nop

;speedzone1	1111111111111111111111111122222222222222222222222222333333333333333333333333334444444444444444444444444455555555555555555555555555
;speedzone2	11111111111111111111111111112222222222222222222222222222333333333333333333333333333344444444444444444444444444445555555555555555555555555555
;speedzone3     111111111111111111111111111111222222222222222222222222222222333333333333333333333333333333444444444444444444444444444444555555555555555555555555555555
;speedzone4     1111111111111111111111111111111122222222222222222222222222222222333333333333333333333333333333334444444444444444444444444444444455555555555555555555555555555555
;               ---1111111111111111111----      ---2222222222222----

;               0         1         2         3         4         5
;               0         0         0         0         0         0
;							^
;							|
;					best read second byte around cycle 40

;data valid at cycle:
;zone1/zone2/zone3/zone4
; 0   / 0   / 0   / 0
; 26  / 28  / 30  / 32
; 52  / 56  / 60  / 64
; 78  / 84  / 90  / 96
; 104 / 112 / 120  / 128

.load_sector_data
		sta .hinibbles,y	;---33333 4
		lax $1c01		;44445555				;read @ 16 / 16 / 16 / 16
		arr #$f0		;44444---				;/!\ ATTENTION, arr potentially fucks the v flag status!
		sta .lonibbles,y	;44444---
		txa			;44445555
		and #%00001111		;----5555
		iny
		sty .lindex+1		;avoid index on lsr

.smc1		bne *			;BRA, y is never 0, could also use bvs, but arr possibly influences V flag /o\
		nop			;so waste either 3,5,7,9 cycles depending on speedzone
		nop
		nop

		ldx $1c01		;56666677				;read @ 40 / 42 / 44 / 46
		cpx #$80		;copy 5 to carry
		rol
		sta .hinibbles,y
		txa			;56666677
		asr #%01111111		;--666667
		sta .lonibbles,y
		iny

.smc2		bne *
		nop
		nop
		nop

		lax $1c01		;77788888				;read @ 67 / 71 / 75 / 79
		and #%00011111
		sta .lonibbles,y	;lda #$0f sax .lonibbles,y if ZP
		txa			;77788888 is fetched again
		ror
.lindex		lsr .lonibbles		;shift out last 7 -> ---66666
		arr #$f0		;77777---
		sta .hinibbles,y
		lda $1c01		;11111222				;read @ 95 / 99 / 103 / 107
		clv								;now finally clear v flag, as we are past the last arr operation
		iny
		lsr

		bvc *
;.smc3		bne *
;		nop
;		nop
;		nop
;		lsr

		sta .hinibbles,y	;-1111122 2
;		nop
;		nop
;		nop
		lda $1c01		;22333334
		clv
		ror			;22233333 4
		sta .lonibbles,y	;store for later processing		;120 cycles min

		bvc *

		and #%00011111		;---33333 4
		iny
		bne .load_sector_data
!if >* != >.load_sector_data { !error "gcr-decoding not in one page: ", .load_sector_data, " - ", * }

		tay
		lda $1c01
		arr #$f0
		tax
		lda .gcr_dec_hi,y
		ora .gcr_dec_lo_s,x
		tay			;checksum

		;checksum/inflate whole sector only if needed
		;either check $100 bytes or just sector len & $fc + 4 bytes depending if a partial (last) or full sector is laoded
		ldx .head_lo + 0
		lda .wanted,x
		ldx #$ff		;blocksize full sector
		cmp .file_size + 1	;as we count up in $100 blocks, if index equals highbyte of size, we have loaded last block
		bne +
		ldx .file_size		;block size last sector (lowbyte filesize)
+
		;remember for later use
		stx .blocksize
		lda #$fc
		sbx #$00		;only eor as much nibbles as necessary (think of partly filled blocks), saves us some time on last block

		;finish gcr inflation and checksum the data
;		ldx #$fc
		;XXX TODO place code on zp and use smc instead of tay/tya.
		;place lo/hinibbles+3 on zp?
-
		lda .lonibbles + 3,x	;4
		lsr .hinibbles + 3,x	;7
		ror			;2
		lsr .hinibbles + 3,x	;7
		arr #$f0		;2
		sta .lonibbles + 3,x	;5

		tya
		ldy .hinibbles + 0,x
		eor .gcr_dec_hi,y
		ldy .lonibbles + 0,x
		eor .gcr_dec_lo_s,y
		ldy .hinibbles + 1,x
		eor .gcr_dec_hi,y
		ldy .lonibbles + 1,x
		eor .gcr_dec_lo,y
		ldy .hinibbles + 2,x
		eor .gcr_dec_hi_s,y
		ldy .lonibbles + 2,x
		eor .gcr_dec_lo,y
		ldy .hinibbles + 3,x
		eor .gcr_dec_hi,y
		ldy .lonibbles + 3,x
		eor .gcr_dec_lo_s,y
		tay

		txa
		sbx #$04
		bcs -

		tya
		;checksum okay?
		bne +

		dey			;y = $ff
		ldx .head_lo + 0
		;fetch blockindex
		lda .wanted,x
		;clear entry in wanted list
		sty .wanted,x
		;done
!if BITFIRE_CONFIG_IN_ORDER = 1 {
		inc .windex
}
		rts
+
		jmp .read_sector

.preamble_size
		jsr .add_preamble
		ldy .blocksize		;set up num of bytes to be transferred
		iny
		tya

.add_preamble
		ldy .pre_cmp
		ldx #$0f
		sax .preamble_lo,y
		lsr
		lsr
		lsr
		lsr
		sta .preamble_hi,y
		inc .pre_cmp
		rts

.send_block
		;send 2 or 4 bytes of blockinfo as preamble
		ldy #$00

		;signal that a block is ready
		lda #BLOCK_READY | BUSY		;DATA_OUT
		sta $1800

		;wait for response of host
		bit $1800
		bpl *-3
		;no need to wait on next part of atn strobe
		;bit $1800
		;bmi *-3
		;on atn going low we have alraedy first data on bus with the upcoming code
.preloop
		ldx .preamble_lo,y
		;we can waste a few cycles here, as the preamble is fetched with several jsr calls on c64 side
		lda .bin2ser,x
		bit $1800
		bmi *-3
		sta $1800		;send first bits on end of atn strobe -> 4th sta on c64 side

		asl
		ora #$10
		ldx .preamble_hi,y	;hinibbles are already shifted into right position
		bit $1800
		bpl *-3
		sta $1800		;second -> first sta $dd02 on c64 side

		lda .bin2ser,x
		iny
		bit $1800
		bmi *-3
		sta $1800		;third -> second sta $dd02 on c64 side

		asl
		ora #$10
		bit $1800
		bpl *-3
		sta $1800		;last -> third sta $dd02 on c64 side
.pre_cmp = * + 1
		cpy #$00
		bcc .preloop

		ldy .blocksize
		;send the data block
		ldx .lonibbles,y
.sendloop
		;this loop is much tighter then the preamble, no cycles to be wasted
		lda .gcr2ser,x
		bit $1800
		bmi *-3
		sta $1800		;read by first lda $dd00 on c64 side

		asl
		ora #$10
		ldx .hinibbles,y
		bit $1800
		bpl *-3
		sta $1800		;read by second lda $dd00 on c64 side

		lda .gcr2ser,x
		dey
		bit $1800
		bmi *-3
		sta $1800		;read by third lda $dd00 on c64 side

		asl
		ora #$10
		ldx .lonibbles,y
		bit $1800
		bpl *-3
		sta $1800		;read by fourth lda $dd00 on c64 side
		cpy #$ff
		bcc .sendloop

		bit $1800
		bmi *-3

		lda #BUSY
		sta $1800

		iny
		;reset size for preamble
		sty .pre_cmp
		;XXX carry is set here, always, might be useful somewhen
		rts

!if >* != >.sendloop { !error "sendloop not in one page" }

		;* = $04ae
.get_byte
		lda #$80

		bit $1800
		bpl *-3

		;free all lines
		sta $1800
		ldx $1800
-
		;did a bit arrive? (bit flip in data in, atn is dropped in same go in first bit)
		cpx $1800
		beq *-3

		;load register
		ldx $1800

		;push bit into carry
		cpx #$04

		;shift in
		ror

		;do until our counter bit ends up in carry
		bcc -

		eor #$ff

		;signal that no block is ready yet and that we are loading, asap
		ldx #BUSY
		stx $1800
		rts

.upload
		;receive end_address of code being uploaded
		;maximum is $0110 - $04ae, buffers at $0500,$0600,$0700 can be used
		;if original code is restored, be sure that get_byte is at the same address

		;receive dest
		jsr .get_byte
		tay
		jsr .get_byte
		sta .dest+1

		;endaddr
		jsr .get_byte
		sta .cmp1
		jsr .get_byte
		sta .cmp2
-
		jsr .get_byte
		sta (.dest),y			;lowbyte of dest is always 0
		iny
		bne +
		inc .dest+1
+
		cpy .cmp1
		bne -
		lda .dest+1
		cmp .cmp2
		bne -
		jmp .drivecode_run
		;register $1800 is not $00 or at least $08 here, this sucks, on second turn it is then $00

;emit warnings only once
!if * > $0500 { !serious "Drivecode is ", * - $0500, " bytes too big!" }
!ifdef .drivecode_end {
	!if $0500 - * < $80 {
		!warn $0500 - *, " bytes remaining for drivecode."
	}
}

;all code form here on might be overwriten upon init of loader, as dir is loaded and both nibble buffers are populated
.drivecode_setup
		;clear wanted list and copy over gcr2ser table
		ldx #$17
-
		lda #$ff
		sta .wanted,x
		lda .gcr2ser_orig,x
		sta .gcr2ser + 8,x
		lda .ht_copy,x
		sta .halftrack_code,x
		lda .mt_copy,x
		sta .motor_code,x
		lda .tu_copy,x
		sta .turn_code,x
		dex
		bpl -

		;generate table for index << 3
		ldx #$09
-
		txa
		asl
		asl
		asl
		tay
		lda .gcr_dec_lo,x
		bmi +		;skip gaps
		sta .gcr_dec_lo_s,y
		lda .gcr_dec_hi,x
		sta .gcr_dec_hi_s,y
		lda .gcr2ser,x
		sta .gcr2ser,y
+
		inx
		tya
		bne -

		ldx #$0f
-
		lda .bin2ser_orig,x
		sta .bin2ser,x
		dex
		bpl -
		lda #18
		sta .track
		jmp .drivecode_run

;code snippets that will be slapped into ZP to save some mem
.ht_copy
!pseudopc .halftrack_code {
.halftrack
		tya
.halftrack_
		bit $1c05
		bmi *-3
.halftrack__
		eor $1c00
		sec
		rol
		and #$03
		eor $1c00
		sta $1c00
		rts
}
.mt_copy
!pseudopc .motor_code {
.motor_on
		lda $1c00
		ora #.VIA2_LED_ON
.motor_on_no_led
		ora #.VIA2_MOTOR_ON
		sta $1c00
		rts
}

.tu_copy
!pseudopc .turn_code {
.turn_disc
		;remember requested side
		sta .temp
		;turn on motor but not LED
		lda $1c00
		and #.VIA2_LED_OFF
		jsr .motor_on_no_led
		jmp .turn_loop
}

.gcr2ser_orig
		!byte $ff,$0e,$0f,$07,$ff,$0a,$0b,$03
		!byte $ff,$ff,$0d,$05,$ff,$00,$09,$01
		!byte $ff,$06,$0c,$04,$ff,$02,$08,$ff
.bin2ser_orig
		!byte %1111, %0111, %1101, %0101, %1011, %0011, %1001, %0001
		!byte %1110, %0110, %1100, %0100, %1010, %0010, %1000, %0000

}

.drivecode_end
.drivecode_size = .drivecode_end - .drivecode_start
