!zone resident {
.resident_include
!src "config.inc"
!convtab pet
!cpu 6510

;loader zp-addresses
.readblock	= $04

;depacker zp-addresses
.lz_match	= $05
.lz_dst		= $07
.lz_bits	= $09

!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
	!if BITFIRE_INCLUDE_DECOMP = 0 {
		!error "Cant't include LOADCOMPD without DECOMP."
	}
}

bitfire_resident_	= *
.open_file
!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		stx .en_dis_bmap		;enable/disable logging in blockmap
}
		inc .skip_la			;transform bcs next_block into lda ($xx),y for the first run
bitfire_send_byte_
		jsr bitfire_set_atn_		;pull atn sets carry for free, wastes cycles so that the floppy can react, no need to wait/poll, assume floppy is idle
		sec				;insert bit as counter
		ror				;fetch first bit and set marker bit
		sta .readblock			;save value
		txa				;-> a = value from $dd02 = $37 set by set_atn

		;will happen in time, no need for that
		;bit $dd00		;wait until floppy is ready (will set all lines to low to allow transfer)
		;bvc *-3
.bit_loop
		and #$20			;preserve bit 5 only
		adc #$2f			;toggle bit 5 and set bit 4 depending on carry -> adc #$30/2f
		ora #$0f			;set all lower bits -> drops also atn on first sent bit
		and #$3f			;leave upper 2 bits untouched!

		sta $dd02
		lsr .readblock			;fetch next bit from filenumber
		bne .bit_loop			;last bit?
						;now reset all lines again
bitfire_reset_atn_
		ldx #$3f			;all lines low again
		top				;skip next two bytes without touching status
bitfire_set_atn_
		ldx #$37
		stx $dd02			;atn high
.ret
		rts

.pollblock
		lda $dd00
		anc #$c0			;focus on bit 7 and 6 and copy bit 7 to carry (set if floppy is idle/eof is reached)
		bne .ret			;only fetch block if both bits are 0 (floppy busy + block_ready)

		;signal that we accept data, by sending an atn strobe
		;XXX TODO reset atn can be omitted? as first bit can be requested by first raise of atn? last bit will then transmitted on atn low?
		jsr bitfire_set_atn_		;trigger send with an atn strobe
		jsr bitfire_reset_atn_		;jsr is used to waste more cycles to give the floppy enough time to react
.skip_la
		;will be changed into a cmp ($xx),y for first run to fall through to the upcoming codeblock else it is a BRA as reset_atn sets carry
		bne .next_block			;BRA

;----------------------------------------------------
;only execute this code here once
;----------------------------------------------------
		;lock out again to make the upcoming code execute once
		dec .skip_la			;$b1 (cmp ($xx),y -> $b0 bne

		;fetch loadaddr hi
		jsr .get_one_byte		;XXX TODO jsr to different routine that either fetches own address or saved address store old value in y and do iny afterwards? then get_one_byte could end after y_val ldy #$xx
!if BITFIRE_INCLUDE_DECOMP = 1 {
	!if BITFIRE_DEPACK_OFFSET = 0 {
bitfire_decomp_addr1
		sta bitfire_lz_sector_ptr1+1
		sta bitfire_lz_sector_ptr2+1
		sta bitfire_lz_sector_ptr3+1
	}
}
bitfire_load_addr1
		sta bitfire_load_addr_hi	;need to save load_addr_hi seperatedly, as the sector_ptr* will be incremented and thus can't be used as load_addr
+
		;fetch loadaddr lo
		jsr .get_one_byte_no
bitfire_load_addr2
!if BITFIRE_DEPACK_OFFSET = 1 {
		clc				;XXX TODO can be omitted?
		adc bitfire_depack_offset_l
}
!if BITFIRE_INCLUDE_DECOMP = 1 {
bitfire_decomp_addr2
		sta bitfire_lz_sector_ptr1
		sta bitfire_lz_sector_ptr2
		sta bitfire_lz_sector_ptr3
}
		sta bitfire_load_addr_lo	;destination lowbyte
!if BITFIRE_DEPACK_OFFSET = 1 {
		lda bitfire_depack_offset_h
		adc bitfire_load_addr_hi
		sta bitfire_load_addr_hi
	!if BITFIRE_INCLUDE_DECOMP = 1 {
bitfire_decomp_addr1
		sta bitfire_lz_sector_ptr1+1
		sta bitfire_lz_sector_ptr2+1
		sta bitfire_lz_sector_ptr3+1
	}
}
+
;----------------------------------------------------

.next_block
		;index
		jsr .get_one_byte
!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		tay			;save index for later use
}
bitfire_load_addr_hi = * + 1
		adc #$00		;no clc needed as get_one_byte returns with carry cleared
		sta bitfire_load_addr_lo + 1
		;XXX TODO save the adc by working with real destination addresses as blockindex? but would require full 32 byte of blockmap? (hibyte)

!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		tya
		;mark block in blockmap, only needed for loadcompd
		jsr .blockmappos
.en_dis_bmap	sta bitfire_blockmap,y
}

		;fetch blocklen
		jsr .get_one_byte_no	;switch already set to rts
bitfire_load_block
		sta .y_val+1

		lda #$a0
		top
.get_one_byte
		lda #$60
		sta .switch
.get_one_byte_no
		ldx #$37
		;XXX TODO set $dd02 to $3f, therefore ommit sty $dd02 at the end?
.getloop
		lda $dd00
		stx $dd02
		lsr
		lsr
		nop
		nop
		ldy #$3f
		ora $dd00
		sty $dd02
		lsr
		lsr
		dec .y_val+1		;waste 6 cycles and decrement
		ora $dd00
		stx $dd02
		lsr
		asr #$fe		;clear carry for free
		sta .nibble
		lda #$c0
		and $dd00
		sty $dd02
.nibble = * +1
		ora #$00
.switch
.y_val		ldy #$00
bitfire_load_addr_lo = * + 1
		sta $4000,y
		bne .getloop

!if >* != >.getloop = 1 { !error "getloop code crosses page!" }
.lz_no_fetch
		rts

!if BITFIRE_INCLUDE_LOADRAW = 1 {
bitfire_loadraw_
!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		ldx #$0c		;disable writing to blockmap, only needed for loadcompd_
}
		jsr .open_file		;easy, open...
-
		jsr .pollblock		;and poll...
		bcc -			;until done
		rts
}

!if BITFIRE_INCLUDE_DECOMP = 1 {
;---------------------------------------------------------------------------------
; REFILL ROUTINES
;---------------------------------------------------------------------------------

.lz_refill_bits
bitfire_lz_sector_ptr1	= *+1
		ldy $beef,x
		sty .lz_bits
		rol .lz_bits
		inx
		bne .lz_no_fetch
.lz_next_sector
		inc bitfire_lz_sector_ptr1+1
		inc bitfire_lz_sector_ptr2+1
		inc bitfire_lz_sector_ptr3+1
!if BITFIRE_INCLUDE_LOADCOMPD != 1 {
.lz_end_of_file
}
.lz_nop_rts
		;finish here when doing a standalone decompress, else have a nop here and continue fetching bytes via loader
		rts			;either rts or nop
.lz_fetch_sector
!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		php
		pha
		;save y, cumbersome :-(
		sty .lz_match
		;z-flag still cleared so skip first pollblock by misusing it for a BRA
		bne +
-
		jsr .pollblock
		bcs ++			;still no data ready, finish
+
		lda .readblock
		jsr .blockmappos	;always returns with cleared carry \o/
		cmp bitfire_blockmap,y	;any change? -> carry set if not
		bcs -			;carry set? then block is not yet loaded. works as buffer is filled from bit 7 to 0
		sta bitfire_blockmap,y	;clear block

		inc .readblock
++
		ldx #$00

		ldy .lz_match
		pla
		plp
		rts

.blockmappos
		ldx #$07
		sbx #$00		;save lower 3 bits in x
		lsr
		lsr
		lsr
		;XXX TODO could limit buffer by using a wraparound like asr #$1e, but when starving loader then too many blocks might be loaded before being depeacked
		tay			;select block by doing >> 3
		lda bitfire_blockmap,y
		eor .bl_bits,x
!if BITFIRE_DEPACK_OFFSET = 0 {
.lz_end_of_file
}
		rts

bitfire_loadcomp_
		ldx #$99		;enable writing to blockmap
		jsr .open_file

		;lda #$00
		;sta .readblock		;start with readpos = 0
		;is always zero as it is shifted empty on open_file

		jsr .lz_fetch_sector	;fetch first block to have some data to work with (depackaddr). Will also set load_addr_lo

		;enable call to lz_maybe_gblock
		ldy #(.lz_maybe_gblock - .lz_skip_en - 2)
		lda #$ea		;nop
		;go!
		bne .loadcompd_entry
}

bitfire_decomp_
		;load_addr is already set by loadraw_/pollblock
		ldx #$00
!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		;handling for the lz_maybe_gblock call
		;disable call to lz_maybe_gblock
		ldy #(.lz_maybe_skip - .lz_skip_en - 2)
		;end lz_fetch_sector in time
		lda #$60		;rts
.loadcompd_entry
		sta .lz_nop_rts
		sty .lz_skip_en + 1
}

;---------------------------------------------------------------------------------
; DECRUNCHER
;---------------------------------------------------------------------------------

		;******** Start the next match/literal run ********

		;always entered with x = 0
.lz_decrunch
		;fetch depack addr
		jsr .lz_refill_bits
		sty .lz_dst-1,x		;x = 1 and x = 2
		cpx #$02
		bne .lz_decrunch
!if BITFIRE_DEPACK_OFFSET = 1 {
		lda .lz_dst
		clc
bitfire_depack_offset_l = * + 1
		adc #$00
		sta .lz_dst
		tya
		;lda .lz_dst + 1
bitfire_depack_offset_h = * + 1
		adc #$00
		sta .lz_dst + 1
		sec
}

		;sec			;This is the main entry point. Forcibly
.lz_type_refill	jsr .lz_refill_bits	;fill up the the bit buffer on entry
.lz_type_check
		bcc .lz_do_match
		beq .lz_type_refill

		;******** Process literal run ********

		lda #$00
-
		rol
		asl .lz_bits
		bne *+5
		jsr .lz_refill_bits
		bcc .lz_lrun_gotten

		asl .lz_bits
		bne -
		jsr .lz_refill_bits
		bne -

.lz_lrun_gotten
		sta .lz_lcopy_len	;Store LSB of run-length
		ldy #$00
.lz_lcopy
bitfire_lz_sector_ptr2	= *+1			;Copy the literal data.
		lda $beef,x
		inx
		bne *+5
		jsr .lz_next_sector
		sta (.lz_dst),y
		iny
.lz_lcopy_len = * + 1
		cpy #$00
		bne .lz_lcopy

		;Time to advance the destination pointer.
		;Maximum run length literals exit here as a type-bit needs
		;to be fetched afterwards
		tya
		beq .lz_maximum		;maximum literal run, bump sector pointers and so on
		clc
		adc .lz_dst
		sta .lz_dst
		bcc *+4
		inc .lz_dst+1

		;One literal run following another only makes sense if the
		;first run is of maximum length and had to be split. As that
		;case has been taken care of we can safely omit the type bit
		;here


		;******** Process match ********

		;check bit -> 1 -> short match (len 2)
		;else long match

.lz_do_match
		lda #$01		;this could be made shorter by using the last bitfetch of the upcoming loop and restoring the carry again by a cmp #$02. Saves bytes, but makes things slower, as eof check is also done with all short matches then

		asl .lz_bits		;first length bit (where a one identifies
		bne *+5			;a two-byte match)
		jsr .lz_refill_bits
		bcc .lz_get_offs	;all done, length is 2, skip further bitfetches (and eof check)
-
		asl .lz_bits
		bne *+5
		jsr .lz_refill_bits
		rol

		asl .lz_bits
		bne *+5
		jsr .lz_refill_bits
		bcc -

		tay
		beq .lz_end_of_file	;A 257-byte (=>$00) run serves as a
					;sentinel
.lz_get_offs
		sta .lz_mcopy_len	;store length at final destination
		lda #%11000000		;fetch 2 more prefix bits
		rol			;previous bit is still in carry \o/
-
		asl .lz_bits
		bne *+5
		jsr .lz_refill_bits
		rol
		bcs -

		tay
		lda .lz_lentab,y
		;eor #$ff		;XXX TODO 8 bits should be represented by $ff instead of $00 if 9 bits are used as well, else things can be made short. table needs to be inverted in that case
		beq .lz_far
-
		asl .lz_bits
		bne *+5
		jsr .lz_refill_bits
		rol
		bcs -
		bmi .lz_short
.lz_far
		eor #$ff		;negate
		tay

bitfire_lz_sector_ptr3	= *+1
		lda $beef,x		;For large offsets we can load the
		inx			;low-byte straight from the stream
		bne .lz_join		;without going throught the shift
		jsr .lz_next_sector	;register
		top
.lz_short
		ldy #$ff
.lz_join
		adc .lz_dst		;subtract offset from lz_dst
		sta .lz_match
		tya			;hibyte
		adc .lz_dst+1
		sta .lz_match+1
		ldy #$ff		;The copy loop. This needs to be run
					;forwards since RLE-style matches can overlap the destination
.lz_mcopy
		iny
		lda (.lz_match),y	;Copy one byte
		sta (.lz_dst),y
.lz_mcopy_len	= *+1
		cpy #$ff
		bne .lz_mcopy

		tya			;Advance destination pointer
;		sec
		adc .lz_dst
		sta .lz_dst
.lz_skip_en
		bcc .lz_maybe_skip
.lz_maximum	inc .lz_dst+1		;This is also used by maximum length

		;only check for new block each time the dst pointer is increased by 256
!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
		bcs .lz_maybe_skip	;always skip, so we only need to manipulate one branch
.lz_maybe_gblock
		stx .lz_match		;save x
		jsr .pollblock		;yes, fetch another block and mark block in bitmap
		ldx .lz_match		;restore x
}
.lz_maybe_skip
					;literals needing an explicit type bit
		asl .lz_bits		;Fetch next type bit
		jmp .lz_type_check

!if BITFIRE_DEPACK_OFFSET = 1 {
.lz_end_of_file
		rts
}
		;******** Fetch some more bits to work with ********

.lz_lentab
		;short offset init values
		!byte %11011111
		!byte %11111011
		!byte %00000000
		!byte %10000000		;same as first byte of bl_bits, could be saved if tables are swapped

		;long offset init values
		!byte %11101111
		!byte %11111101
		!byte %10000000
		!byte %11110000

!if BITFIRE_INCLUDE_LOADCOMPD = 1 {
.bl_bits
		!byte $80,$40,$20,$10,$08,$04,$02,$01

!if BITFIRE_BLOCKMAP_ADDRESS = 0 {
bitfire_blockmap
		;XXX TODO theoretically we can only load from $2000-$d000 in a demo ($b0 blocks) so we would max need a blockbuffer of $16 bytes
		!byte $00,$00,$00,$00,$00,$00,$00,$00
		!byte $00,$00,$00,$00,$00,$00,$00,$00
		!byte $00,$00,$00,$00,$00,$00,$00,$00
		!byte $00,$00;,$00,$00,$00,$00;,$00,$00
} else {
bitfire_blockmap = BITFIRE_BLOCKMAP_ADDRESS
}
}
}	;endif DECOMP
}	;!zone resident




;		ldy #$3f
;		lda $dd00
;		stx $dd02
;		lsr
;		ora $dd00
;		sty $dd02
;		lsr
;		sta data
;		lda $dd00
;		stx $dd02
;		and #$88
;		ora data
;		lsr
;		sta data
;		lda $dd00
;		sty $dd02
;		and #$88
;		ora data
;
;		ldy #$00
;		sta $0400,y
;		dey
;		bne -

;.sendloop
;		;this loop is much tighter then the preamble, no cycles to be wasted
;		ldx .lonibbles,y
;		lda .gcr2ser,x
;		bit $1800
;		bmi *-3
;		sta $1800		;read by first lda $dd00 on c64 side
;
;		asl
;		ora #$10
;		ldx .hinibbles,y
;		bit $1800
;		bpl *-3
;		sta $1800		;read by second lda $dd00 on c64 side
;
;		lda .gcr2ser,x
;		dey
;		bit $1800
;		bmi *-3
;		sta $1800		;read by third lda $dd00 on c64 side
;
;		asl
;		ora #$10
;		bit $1800
;		bpl *-3
;		sta $1800		;read by fourth lda $dd00 on c64 side
;		cpy #$ff
;		bcc .sendloop
;
;		bit $1800
